import os
from pickle import load

from dill import loads, dumps

import random
import string

import inspect

from pipeline.schemas import (
    PipelineVariableSchema,
    PipelineFunctionSchema,
    PipelineGraphNodeSchema,
    PipelineGraph,
)

CACHE_DIR = os.getenv("PIPELINE_CACHE_DIR", "./cache")

if not os.path.exists(CACHE_DIR):
    os.makedirs(CACHE_DIR)
elif not os.path.isdir(CACHE_DIR):
    raise Exception("Cache dir '%s' is not a valid dir." % CACHE_DIR)


class Variable(object):
    schema: PipelineVariableSchema

    def __init__(self, **schema):
        if Pipeline._current_pipeline_defining:
            self.schema = PipelineVariableSchema(**schema)
            Pipeline.add_variable(self.schema)


class Pipeline(object):
    defined_pipelines = {}

    _current_pipeline: PipelineGraph = None
    _current_pipeline_defining = False

    def __init__(self, pipeline_name):
        self.pipeline_name = pipeline_name

    # __enter__ - called at the end of a "with" block.
    def __enter__(self):

        Pipeline._current_pipeline = PipelineGraph(
            inputs=[],
            outputs=[],
            variables=[],
            graph_nodes=[],
            models=[],
            name=self.pipeline_name,
        )

        Pipeline._current_pipeline_defining = True
        return self

    # __exit__ - called at the end of a "with" block.
    def __exit__(self, type, value, traceback):
        Pipeline.defined_pipelines[self.pipeline_name] = self._current_pipeline.copy()

        Pipeline._current_pipeline_defining = False

    def output(self, *outputs):
        for _output in outputs:
            if not isinstance(_output, PipelineVariableSchema):
                raise Exception("Can only return PipelineVariables idiot.")

            variable_index = self._current_pipeline.variables.index(_output)
            if variable_index != -1:

                self._current_pipeline.variables[variable_index].is_output = True

                for node in self._current_pipeline.graph_nodes:
                    for op in node.outputs:
                        if op.variable_name == _output.variable_name:
                            op.is_output = True

                """Pipeline._current_pipeline.outputs.append(
                    PipelineOutputVariableSchema(
                        variable=self._current_pipeline.variables[variable_index]
                    )
                )"""

    def save(self, dir, **kwargs):
        return self._current_pipeline.save(dir, **kwargs)

    @staticmethod
    def load(path):
        # with open(os.path.join(path, "config.json"), "r") as config_file:
        Pipeline._current_pipeline = PipelineGraph.parse_file(
            os.path.join(path, "config.json")
        )

        # Load functions in graph nodes
        for node in Pipeline._current_pipeline.graph_nodes:
            function_file_name = node.pipeline_function.function_file_name
            with open(os.path.join(path, function_file_name), "rb") as function_file:
                function_data = loads(function_file.read())
                node.pipeline_function.function = function_data
            if node.pipeline_function.bound_class_file_name != None:
                with open(
                    os.path.join(path, node.pipeline_function.bound_class_file_name),
                    "rb",
                ) as cls_file:
                    cls_data = loads(cls_file.read())
                    node.pipeline_function.bound_class = cls_data

        # Load models
        for model in Pipeline._current_pipeline.models:
            with open(
                os.path.join(path, model.model_file_name),
                "rb",
            ) as model_file:
                model_data = loads(model_file.read())
                model.model = model_data
        # Load variables
        for variable in Pipeline._current_pipeline.variables:
            with open(
                os.path.join(path, variable.variable_type_file_path),
                "rb",
            ) as variable_type_file:
                variable_type_data = loads(variable_type_file.read())
                variable.variable_type = variable_type_data

        return Pipeline._current_pipeline

    # def save(path):

    @staticmethod
    def add_variable(new_variable: PipelineVariableSchema):

        if Pipeline._current_pipeline_defining:

            Pipeline._current_pipeline.variables.append(new_variable)

            """if new_variable.is_input:
                Pipeline._current_pipeline.inputs.append(
                    PipelineInputVariableSchema(variable=new_variable)
                )
            elif new_variable.is_output:
                Pipeline._current_pipeline.outputs.append(
                    PipelineOutputVariableSchema(variable=new_variable)
                )"""
        else:
            raise Exception("Cant add a variable when not defining a pipeline!")

    @staticmethod
    def get_pipeline(pipeline_name) -> PipelineGraph:
        return Pipeline.defined_pipelines[pipeline_name]


def pipeline_function(function):
    def execute_func(*args, **kwargs):

        if not Pipeline._current_pipeline_defining:
            print("Call")
            return function(*args, **kwargs)
        else:

            function_ios = function.__annotations__
            if not "return" in function_ios:
                raise Exception(
                    "Must include an output type e.g. 'def my_func(...) -> int:'"
                )

            processed_args = []

            for input_arg in args:
                if isinstance(input_arg, Variable):
                    if not input_arg.schema in Pipeline._current_pipeline.variables:
                        raise Exception(
                            "Vairble not found, have you forgotten to define it as in input? "
                        )
                    processed_args.append(input_arg.schema)
                elif hasattr(input_arg, "__pipeline_model__"):
                    if function.__pipeline_function__.bound_class == None:
                        function.__pipeline_function__.bound_class = input_arg
                elif isinstance(input_arg, PipelineVariableSchema):
                    print(input_arg)
                    if not input_arg in Pipeline._current_pipeline.variables:
                        raise Exception(
                            "Vairble not found, have you forgotten to define it as in input? "
                        )
                    processed_args.append(input_arg)
                else:
                    raise Exception(
                        "You can't input random variables, follow the way of the Pipeline. Got type"
                    )

            node_output = PipelineVariableSchema(
                variable_type=function.__annotations__["return"]
            )
            Pipeline.add_variable(node_output)

            # Everytime this function is called we have to add a new node in the graph
            graph_node = PipelineGraphNodeSchema(
                pipeline_function=function.__pipeline_function__,
                inputs=processed_args,
                outputs=[node_output],
            )
            Pipeline._current_pipeline.graph_nodes.append(graph_node)

            return graph_node.outputs[0]

    function_inputs = {
        function_i: function.__annotations__[function_i]
        for function_i in function.__annotations__
        if not function_i == "return"
    }

    function.__pipeline_function__ = PipelineFunctionSchema(
        inputs=function_inputs,
        name=function.__name__,
        function=function,
    )

    return execute_func



import os
import string
import random
import inspect

from dill import dumps

from hashlib import sha256

from pydantic import BaseModel

from typing import List, Optional, Callable, Any


class PipelineVariableSchema(BaseModel):
    variable_type: Any
    variable_type_file_path: Optional[str]
    variable_name: str
    is_input: bool = False
    is_output: bool = False

    def __init__(self, **kwargs):
        if not "variable_name" in kwargs:
            kwargs["variable_name"] = "".join(
                random.choice(string.ascii_lowercase) for i in range(10)
            )
        super().__init__(**kwargs)

    def json(self):
        # print("coming in")
        return self.dict()

    def dict(self, *args, **kwargs):
        # print("Things")
        return {
            "variable_type": str(self.variable_type.__name__),
            "variable_type_file_path": self.variable_type_file_path,
            "variable_name": self.variable_name,
            "is_input": self.is_input,
            "is_output": self.is_output,
        }


class PipelineModel(BaseModel):
    model: Any
    name: str
    hash: Optional[str]
    model_file_name: Optional[str]

    def __init__(self, *args, **kwargs):
        if not "name" in kwargs:
            kwargs["name"] = "".join(
                random.choice(string.ascii_lowercase) for i in range(10)
            )
        super().__init__(**kwargs)

    def dict(self, *args, **kwargs):
        return {
            "name": self.name,
            "hash": self.hash,
            "model_file_name": self.model_file_name,
        }


class PipelineFunctionSchema(BaseModel):
    inputs: dict
    name: str
    hash: Optional[str]
    function: Optional[Callable]
    function_file_name: Optional[str]
    bound_class: Optional[Any]
    bound_class_file_name: Optional[str]

    def dict(self, *args, **kwargs):
        return {
            "inputs": {
                input_name: self.inputs[input_name].__class__.__name__
                for input_name in self.inputs
            },
            "name": self.name,
            "hash": self.hash,
            "function_file_name": self.function_file_name,
            "bound_class_file_name": self.bound_class_file_name,
        }


class PipelineGraphNodeSchema(BaseModel):
    id: str
    pipeline_function: PipelineFunctionSchema
    inputs: List[PipelineVariableSchema]
    outputs: List[PipelineVariableSchema]

    def __init__(self, *args, **kwargs):
        if not "id" in kwargs:
            kwargs["id"] = "pipeline_graph_node_" + "".join(
                random.choice(string.ascii_lowercase) for i in range(20)
            )
        super().__init__(*args, **kwargs)


class PipelineGraph(BaseModel):
    name: str
    variables: List[PipelineVariableSchema]
    graph_nodes: List[PipelineGraphNodeSchema]
    models: List[PipelineModel]

    def run(self, *inputs):

        input_variables: List[PipelineVariableSchema] = [
            var for var in self.variables if var.is_input
        ]
        output_variables: List[PipelineVariableSchema] = [
            var for var in self.variables if var.is_output
        ]

        for model in self.models:
            if hasattr(model.model, "load"):
                model.model.load(None)

        if len(inputs) != len(input_variables):
            raise Exception(
                "Mismatch of number of inputs, expecting %u got %s"
                % (len(input_variables), len(inputs))
            )
        running_variables = {}
        for i, input in enumerate(inputs):
            if not isinstance(input, input_variables[i].variable_type):
                raise Exception(
                    "Input type mismatch, expceted %s got %s"
                    % (
                        input_variables[i].variable_type,
                        input.__class__,
                    )
                )
            running_variables[input_variables[i].variable_name] = input

        for node in self.graph_nodes:
            node_inputs = node.inputs
            node_function = node.pipeline_function
            node_output = node.outputs[0]

            function_inputs = []
            for _input in node_inputs:
                function_inputs.append(running_variables[_input.variable_name])

            if node_function.bound_class != None:
                output = node_function.function(
                    node_function.bound_class, *function_inputs
                )
            else:
                output = node_function.function(*function_inputs)

            running_variables[node_output.variable_name] = output

        return_variables = []
        for output_variable in output_variables:
            return_variables.append(running_variables[output_variable.variable_name])

        return return_variables

    def save(self, save_dir, tar=False, overwrite=True):

        base_save_path = os.path.join(save_dir, "%s" % self.name)
        if os.path.exists(base_save_path):
            if not overwrite and len(os.listdir(base_save_path)) > 0:
                raise Exception("Save directory not empty.")
            elif overwrite and len(os.listdir(base_save_path)) > 0:
                for file in os.listdir(base_save_path):
                    os.remove(os.path.join(base_save_path, file))

        print("Saving pipeline in: '%s'" % base_save_path)
        os.makedirs(base_save_path, exist_ok=True)

        # Save functions
        for node in self.graph_nodes:
            pipeline_function = node.pipeline_function
            function_data = dumps(pipeline_function.function)
            function_source = inspect.getsource(pipeline_function.function)
            function_hash = sha256(function_source.encode()).hexdigest()
            save_file_name = function_hash + ".fn"

            # Update node info
            node.pipeline_function.hash = function_hash
            node.pipeline_function.function_file_name = save_file_name

            with open(
                os.path.join(base_save_path, save_file_name), "wb"
            ) as node_function_file:
                node_function_file.write(function_data)
            if node.pipeline_function.bound_class != None:
                bound_class_data = dumps(node.pipeline_function.bound_class)
                # bound_class_source = inspect.getsource(
                #    node.pipeline_function.bound_class
                # )
                bound_class_hash = sha256(bound_class_data).hexdigest()
                bound_class_file_name = bound_class_hash + ".cls"
                node.pipeline_function.bound_class_file_name = bound_class_file_name
                with open(
                    os.path.join(base_save_path, bound_class_file_name), "wb"
                ) as bound_class_file:
                    bound_class_file.write(bound_class_data)
        # Save models
        for model in self.models:
            model_data = dumps(model.model)
            model_hash = sha256(model_data).hexdigest()
            model.hash = model_hash
            model_file_name = model_hash + ".mdl"
            model.model_file_name = model_file_name
            with open(
                os.path.join(base_save_path, model_file_name), "wb"
            ) as model_file:
                model_file.write(model_data)

        # Save variables
        for variable in self.variables:
            type_data = dumps(variable.variable_type)
            type_hash = sha256(type_data).hexdigest()
            type_file_name = type_hash + ".typ"
            variable.variable_type_file_path = type_file_name
            with open(os.path.join(base_save_path, type_file_name), "wb") as type_file:
                type_file.write(type_data)

        with open(os.path.join(base_save_path, "config.json"), "w") as config_file:

            config_file.write(self.json())



from pipeline import Pipeline, CACHE_DIR
from pipeline.schemas import PipelineModel
import tarfile


class pipeline_model(object):
    def __init__(
        self, model_class=None, *, file_or_dir: str = None, compress_tar=False
    ):
        if model_class != None:
            model_class.__pipeline_model__ = True

        self.compress_tar = compress_tar
        self.model_class = model_class
        self.file_or_dir = file_or_dir

    def __call__(self, *args, **kwargs):

        if len(args) + len(kwargs) == 1:
            self.model_class = args[0]
            self.model_class.__pipeline_model__ = True
            return self.__function_exe__
        else:
            print(len(args) + len(kwargs))
            return self.__function_exe__(*args, **kwargs)

    def __function_exe__(self, *args, **kwargs):
        if not Pipeline._current_pipeline_defining:
            return self.model_class(*args, **kwargs)
        else:
            created_model = self.model_class(*args, **kwargs)
            model_schema = PipelineModel(model=created_model)
            Pipeline._current_pipeline.models.append(model_schema)

            model_functions = [
                model_attr
                for model_attr in dir(created_model)
                if callable(getattr(created_model, model_attr))
                and model_attr[:2] != "__"
            ]
            return created_model



from pipeline import pipeline_function
from pipeline.model import pipeline_model

from transformers import AutoModelForCausalLM, AutoTokenizer


@pipeline_model()
class TransformersModel(object):
    model: AutoModelForCausalLM = None
    tokenizer: AutoTokenizer = None

    def __init__(self, model_path, tokenizer_path):
        self.model_path = model_path
        self.tokenizer_path = tokenizer_path

    @pipeline_function
    def predict(self, input: str, **kwargs: dict) -> str:
        input_ids = self.tokenizer(input, return_tensors="pt").input_ids
        output_tokens = self.model.generate(input_ids, **kwargs)
        output_str = self.tokenizer.batch_decode(output_tokens)[0]
        return output_str

    @pipeline_function
    def load(self, path: str) -> None:
        self.model = AutoModelForCausalLM.from_pretrained(self.model_path)
        self.tokenizer = AutoTokenizer.from_pretrained(self.tokenizer_path)
